package com.me.common.ecs;

import com.badlogic.gdx.utils.Array;
import com.me.common.ecs.event.Event;
import com.me.common.ecs.event.Listener;

import java.util.HashMap;
import java.util.Map;

public class Engine {

    private ComponentBag[] components;
    protected Array<BaseSystem> systems;

    private EntityManager entityManager;
    private ListenerRegistry listenerRegistry;

    private Map<Class<? extends Component>, ComponentMapper> componentMappers;

    public Engine() {
        this.entityManager = new EntityManager(this);

        this.systems = new Array<>();
        this.listenerRegistry = new ListenerRegistry();
        this.componentMappers = new HashMap<>();
    }

    public void registerComponentClass(Class<? extends Component> clazz) {
        ComponentType.registerComponentType(clazz);
    }

    public void registerSystem(BaseSystem system) {
        this.systems.add(system);
    }

    public void registerListener(Listener listener) {
        listenerRegistry.registerListener(listener);
    }

    public void ready() {
        this.components = new ComponentBag[ComponentType.getRegisteredComponentTypeCount()];
        for (int i = 0; i < components.length; i++) {
            components[i] = new ComponentBag();
        }
    }

    public void update(float dt) {
        entityManager.update();

        for (BaseSystem system : systems) {
            system.preProcess();
            system.process(dt);
            system.postProcess();
        }
    }

    public Array<Entity> getEntities() {
        return entityManager.entities;
    }

    public int getEntityCount() {
        return entityManager.entities.size;
    }

    public Entity createEntity() {
        return entityManager.create();
    }

    public void callEvent(Event event) {
        listenerRegistry.callEvent(event);
    }

    protected void refreshEntity(Entity entity) {
        entityManager.queueRefresh(entity);
    }

    protected void removeAllEntityComponents(int entityId) {
        for (int i = 0; i < components.length; i++) {
            components[i].insert(entityId, null);
        }
    }

    protected void addEntityComponent(Entity entity, Component component) {
        ComponentType type = ComponentType.getComponentType(component.getClass());
        components[type.id].insert(entity.id, component);
        entity.addComponentType(type);
    }


    protected void removeEntityComponent(Entity entity, Component component) {
        ComponentType type = ComponentType.getComponentType(component.getClass());
        components[type.id].remove(entity.id);
        entity.removeComponentType(type);
    }

    protected <T extends Component> T getEntityComponent(Entity entity, Class<T> clazz) {
        return clazz.cast(components[ComponentType.getTypeId(clazz)].get(entity.id));
    }

    protected Component getEntityComponent(Entity entity, ComponentType type) {
        return components[type.id].get(entity.id);
    }

    @SuppressWarnings("unchecked")
    public <T extends Component> ComponentMapper<T> getComponentMapper(Class<T> typeClass) {
        ComponentMapper<T> mapper = componentMappers.get(typeClass);
        if (mapper == null) {
            mapper = new ComponentMapper<>(this, typeClass);
            componentMappers.put(typeClass, mapper);
        }
        return mapper;
    }

}
